# coding=utf-8
# Command Line Parser
from __future__ import print_function, unicode_literals

import re
import sys
from collections import OrderedDict

from bgmi import __version__
from bgmi.utils.utils import unicodeize

HELP = ('-h', '--help')
JUST_GROUP = 0
SUB_PARSER = 1


def _error(message):
    sys.stderr.write('error: %s, try \'-h\' for help\n' % message)
    raise SystemExit(1)


class NameSpace(object):
    NameSpace_Action_Name = None

    def __init__(self):
        self.__dict__['__value'] = {}

    def __setattr__(self, key, value):
        if key == 'NameSpace_Action_Name':
            self.__dict__['NameSpace_Action_Name'] = value
        else:
            self.__value[key] = value

    def __getattr__(self, item):
        if item == '_NameSpace__value':
            return self.__dict__['__value']

        if item in self.__dict__['__value']:
            return self.__value[item]
        else:
            return self.__dict__[item]

    def __getitem__(self, item):
        return self.__getattr__(item)

    def __str__(self):
        return 'Namespace(\'%s\')' % str(self.__value)

    def __repr__(self):
        return '\'%s\'' % self.__str__()

    def __eq__(self, other):
        return other == self.NameSpace_Action_Name


class _CommandParserMixin(object):

    def add_argument(self, name, dest=None, arg_type=None, required=False, choice=None, default=None, mutex=None,
                     call=False, help='', hidden=False):
        # self.container._check_conflict(name)
        self._check_conflict(name)
        argument = Argument(name=name, dest=dest, arg_type=arg_type, choice=choice,
                            default=default, call=call, help=help, group=self, required=required,
                            hidden=hidden)
        self.arguments.update({name: argument})

        # update mutex options dict
        if mutex is not None:
            self.mutex.update({name: mutex, mutex: name})

    def add_arg_group(self, name, help=''):
        self._check_group_conflict(name)
        group = ArgumentGroup(name=name, help=help, container=self)
        self.argument_groups.update({name: group})
        return group

    def _check_conflict(self, name):
        if name in self.arguments:
            _error('conflict argument name: %s' % name)

    def _check_group_conflict(self, name):
        if name in self.argument_groups:
            _error('conflict argument group name: %s' % name)

    def _check_group_name(self, name):
        if re.match('^([a-zA-Z0-9_]+)$', name) is None or re.match('^[0-9]', name):
            _error('invalid argument_name \'%s\'' % name)

    def _get_args(self, arg, _sys_args):
        '''Get argument(s) by arg_type
        '''
        if arg.type == '+':
            # One or more arguments
            if not _sys_args or _sys_args[-1].startswith('-'):
                _error('%s expected at least one argument' % arg.name)
            _arg_list = []
            while _sys_args and not _sys_args[-1].startswith('-'):
                _arg_list.append(_sys_args.pop())
            return _arg_list
        elif arg.type == '*':
            # Zero or more arguments
            raise NotImplementedError
        elif arg.type == '1':
            # Just one argument
            if not _sys_args:
                _error('%s expected one argument' % arg.name)
            return _sys_args.pop()
        elif arg.type is None:
            # None
            return True

    def _set_default(self):
        '''Set default value for each argument
        '''
        for arg in self.arguments.values():
            if arg.default is not None:
                value = arg.default
            else:
                value = None

            setattr(self.namespace, arg.dest, value)

        for group in self.argument_groups.values():
            group._set_default()
            setattr(self.namespace, group.name, group.namespace)

    def _get_positional_args(self):
        self.positional_args = [i for i in self.arguments.values()
                                if not i.optional and not i.type == 's']
        self._optional_args = [i for i in self.arguments.values()
                               if i not in self.positional_args and not i.type == 's']

        self._positional_args = self.positional_args[::-1]

    def _parse_command(self, parser_group, _sys_args_list):
        while _sys_args_list:
            arg = _sys_args_list.pop()
            _mutex_arg = parser_group.mutex.get(arg, None)

            if _mutex_arg is not None:
                if arg in parser_group._mutex_list:
                    _error('arguments %s and %s are mutually exclusive' % (arg, _mutex_arg))
                parser_group._mutex_list.append(_mutex_arg)

            if arg in HELP:
                parser_group.container.print_help()

            if arg.startswith('-'):
                arg_instance = parser_group.arguments.get(arg, None)
                if arg_instance is None:
                    if not parser_group.argument_groups:
                        _error('unrecognized arguments: %s' % arg)
                    _sys_args_list.append(arg)
                    break
                else:
                    value = self._get_args(arg_instance, _sys_args_list)
            else:
                if parser_group._positional_args:
                    arg_instance = parser_group._positional_args.pop()
                else:
                    if not parser_group.argument_groups:
                        _error('unrecognized arguments: %s' % arg)
                    _sys_args_list.append(arg)
                    break

                if arg_instance.type is None:
                    value = arg
                elif arg_instance.type == '+':
                    value = [arg]
                    for i in _sys_args_list[::-1]:
                        if not i.startswith('-'):
                            value.append(_sys_args_list.pop())
                        else:
                            break
                else:
                    _error('unsupported arg type: %s' % arg.type)

            if arg_instance.choice:
                if value not in map(str, arg_instance.choice):
                    _error('unexpected choice of %s (%s): %s' % (arg_instance.name,
                                                                 ', '.join(map(str, arg_instance.choice)), value))

            setattr(parser_group.namespace, arg_instance.dest, value)

        for arg in parser_group.arguments.values():
            if arg.required and parser_group.namespace.__dict__['__value'].get(arg.dest) is None:
                _error('argument \'%s\' is required' % arg.name)

        if isinstance(parser_group, ArgumentGroup):
            setattr(parser_group.container.namespace, parser_group.name, parser_group.namespace)


class ArgumentGroup(_CommandParserMixin):
    def __init__(self, name, help='', container=None, type=0):
        self.namespace = NameSpace()
        self.arguments = OrderedDict()
        self.name = name
        self.help = help
        self.container = container
        self.sub = self.argument_groups = OrderedDict()
        self.mutex = {}
        self._mutex_list = []
        # 0 - 普通的 group 我们普通的摇
        # 1 - sub parser group
        self.type = type

    def add_sub_parser(self, name, help=None):
        self._check_group_name(name)
        if name in self.argument_groups:
            _error('conflict sub parser name: %s' % name)
        sub_parser_group = ArgumentGroup(name=name, help=help, container=self, type=1)
        self.argument_groups.update({name: sub_parser_group})
        return sub_parser_group

    def parse_command(self, _sys_args_list):
        self._get_positional_args()
        self._set_default()

        self._parse_command(self, _sys_args_list)

        while _sys_args_list:
            arg = _sys_args_list.pop()

            sub_parser = self.argument_groups.get(arg, None)
            if not sub_parser:
                for group in self.argument_groups.values():
                    if arg in group.argument_groups:
                        sub_parser = group.argument_groups.get(arg)
                        break

            if sub_parser:
                sub_parser.parse_command(_sys_args_list)
                self.namespace.NameSpace_Action_Name = sub_parser.name
                break
            else:
                _error('unrecognized arguments: %s' % arg)

        return self.namespace

    def print_help(self):
        return self.container.print_help()

    def __str__(self):
        return 'ArgumentGroup:<%s>' % self.name

    def __repr__(self):
        return 'ArgumentGroup:<%s>' % self.name


class Argument(object):
    args_type_list = ('+', '*', '1', None)

    def __init__(self, name, dest, arg_type=None, choice=None, required=False, default=None, call=False,
                 help='', group=None, hidden=False):
        # check the validity of arg_type
        if arg_type not in self.args_type_list:
            _error('unexpected args type: %s' % arg_type)

        # positional arguments' argument type should be None
        if not name.startswith('-') and arg_type not in ('+', None):
            _error('unexpected positional argument type: %s' % arg_type)

        if not isinstance(choice, (list, tuple, set, str, type(None), )):
            _error('unexpected choice type: %s' % type(choice))

        # if dest not set, use name as dest
        if dest is None:
            if name.startswith('-'):
                _temp = name
                while _temp.startswith('-'):
                    _temp = _temp[1:]
                dest = _temp
            else:
                dest = name
            dest = dest.replace('-', '_')

        if arg_type == '*':
            raise NotImplementedError('not implemented arg type: *')

        self.name = name
        self.dest = dest
        self.type = arg_type
        self.call = call
        self.help = help
        self.default = default
        self.group = group
        self.choice = choice
        self.required = required
        self.hidden = hidden
        self.optional = self.name.startswith('-')

    def __str__(self):
        return 'Argument<\'%s\'>' % self.name

    def __repr__(self):
        return 'Argument<\'%s\'>' % self.name


class CommandParser(_CommandParserMixin):
    sub = None
    name = None

    def __init__(self):
        self.namespace = NameSpace()
        self._positional_args = []
        self._optional_args = []
        self.arguments = OrderedDict()
        self.argument_groups = OrderedDict()
        self.mutex = {}
        self._mutex_list = []
        self.container = self

    def parse_command(self):
        sys_argv = list(map(unicodeize, sys.argv))
        self.sys_args = self._sys_args = sys_argv[1:][::-1]
        self._get_positional_args()
        self._set_default()

        self._parse_command(self, self._sys_args)

        for group in self.argument_groups.values():
            group.parse_command(self._sys_args)

        return self.namespace

    def print_help(self):

        def search(name, group):
            if name in group.argument_groups:
                return group.argument_groups.get(name)
            else:
                for group in group.argument_groups.values():
                    if name in group.argument_groups:
                        return group.argument_groups.get(name)

        def get_arg_form(arg):
            if not arg.hidden:
                if arg.optional:
                    if arg.type == '+':
                        usage = '%s [%s ...]' % (arg.name, arg.dest.upper())
                    elif arg.type == '1':
                        usage = '%s %s' % (arg.name, arg.dest.upper())
                    elif arg.type is None:
                        usage = arg.name
                    else:
                        usage = ''
                else:
                    if arg.type == '+':
                        usage = '[%s] ... ' % arg.name
                    elif arg.type is None:
                        usage = '<%s> ' % arg.name
                    else:
                        usage = ''
                return usage
            else:
                return None

        def print_group_help(group):
            group._get_positional_args()
            if not group.positional_args and not group._optional_args and not group.sub:
                return

            group_name = group.name.title() if not group.name.startswith('-') else group.name
            sys.stdout.write('\n%s:\n' % group_name)

            for arg in group.arguments.values():
                usage = get_arg_form(arg)
                if usage:
                    sys.stdout.write('  %-28s%s\n' % (arg.name, arg.help))

            for sub in group.sub.values():
                sys.stdout.write('  %-28s%s\n' % (sub.name, sub.help))

        def _print_help(container, usage):
            container._get_positional_args()

            if container._positional_args:
                for arg in list(container._positional_args)[::-1]:
                    if not arg.hidden:
                        usage += get_arg_form(arg)

            if container._optional_args:
                usage += '[options] '

            if container.argument_groups:
                groups = [group.name for group in container.argument_groups.values() if group.type == JUST_GROUP]
                for group in container.argument_groups.values():
                    if group.type == SUB_PARSER:
                        groups.append('sub_commands')
                        break
                usage += '<{0}>'.format('|'.join(groups))

            sys.stdout.write('%s \n' % usage)

            if container.argument_groups:
                help_text = ''
                for group in list(container.argument_groups.values())[::-1]:
                    if group.type == SUB_PARSER:
                        help_text += '  %-28s%s\n' % (group.name, group.help)
                if help_text:
                    sys.stdout.write('\nSub Commands:\n')
                    sys.stdout.write(help_text)

                for group in list(container.argument_groups.values())[::-1]:
                    if group.type == JUST_GROUP:
                        print_group_help(group)

            if container._positional_args:
                sys.stdout.write('\nArguments: \n')
                for arg in container._positional_args[::-1]:
                    sys.stdout.write('  %-28s%s\n' % (arg.name, arg.help))

            if isinstance(container, ArgumentGroup):
                if container._optional_args:
                    sys.stdout.write('\nOptions: \n')

                for arg in container._optional_args:
                    form = get_arg_form(arg)
                    if form:
                        sys.stdout.write('  %-28s%s\n' % (form, arg.help))

            if self._optional_args:
                sys.stdout.write('\nGeneral Options: \n')

            for arg in self._optional_args:
                form = get_arg_form(arg)
                if form:
                    sys.stdout.write('  %-28s%s\n' % (form, arg.help))

        self.sys_args = sys.argv[1:][::-1]
        self._sys_args = sys.argv[1:]
        sys.stdout.write('BGmi %s (https://github.com/RicterZ/BGmi)\n\nUsage: \n' % __version__)

        arg_list = []
        while self.sys_args:
            arg = self.sys_args.pop()
            if arg not in HELP:
                arg_list.append(arg)
                break
        else:
            arg = None

        ret = search(arg, self)
        while ret and self.sys_args:
            arg = self.sys_args.pop()
            if arg not in HELP:
                _ = search(arg, ret)
                if _:
                    ret = _
                    arg_list.append(arg)

        usage = '  bgmi '
        if isinstance(ret, ArgumentGroup):
            usage += ' '.join(arg_list)
            usage += ' '
            _print_help(ret, usage)
        else:
            _print_help(self, usage)

        exit(0)



# coding=utf-8
from __future__ import print_function, unicode_literals
import os
import sys
import locale
import codecs
import signal
import sqlite3
import time
import platform

import bgmi.config
from bgmi.command import CommandParser
from bgmi.config import BGMI_PATH, DB_PATH, write_config
from bgmi.download import download_prepare
from bgmi.fetch import fetch, bangumi_calendar, get_maximum_episode
from bgmi.models import Bangumi, Followed, Download, Filter, STATUS_FOLLOWED, STATUS_UPDATED,\
    STATUS_NORMAL, STATUS_NOT_DOWNLOAD
from bgmi.sql import CREATE_TABLE_BANGUMI, CREATE_TABLE_FOLLOWED, CREATE_TABLE_DOWNLOAD, CREATE_TABLE_FOLLOWED_FILTER
from bgmi.utils.utils import print_warning, print_info, print_success, print_error, print_version
from bgmi.download import get_download_class


# Wrap sys.stdout into a StreamWriter to allow writing unicode.
if bgmi.config.IS_PYTHON3:
    file_ = sys.stdout.buffer
    sys.stdout = codecs.getwriter(locale.getpreferredencoding())(file_)
else:
    reload(sys)
    sys.setdefaultencoding('utf-8')
    input = raw_input


ACTION_ADD = 'add'
ACTION_FETCH = 'fetch'
ACTION_FILTER = 'filter'
ACTION_DELETE = 'delete'
ACTION_UPDATE = 'update'
ACTION_CAL = 'cal'
ACTION_CONFIG = 'config'
ACTION_DOWNLOAD = 'download'
ACTION_FOLLOWED = 'followed'
ACTION_MARK = 'mark'
ACTIONS = (ACTION_ADD, ACTION_DELETE, ACTION_UPDATE, ACTION_CAL,
           ACTION_CONFIG, ACTION_FILTER, ACTION_FETCH, ACTION_DOWNLOAD, ACTION_FOLLOWED,
           ACTION_MARK, )

FILTER_CHOICE_TODAY = 'today'
FILTER_CHOICE_ALL = 'all'
FILTER_CHOICE_FOLLOWED = 'followed'
FILTER_CHOICES = (FILTER_CHOICE_ALL, FILTER_CHOICE_FOLLOWED, FILTER_CHOICE_TODAY)


DOWNLOAD_ACTION_LIST = 'list'
DOWNLOAD_ACTION_MARK = 'mark'
DOWNLOAD_ACTION = (DOWNLOAD_ACTION_LIST, DOWNLOAD_ACTION_MARK)


DOWNLOAD_CHOICE_LIST_ALL = 'all'
DOWNLOAD_CHOICE_LIST_NOT_DOWNLOAD = 'not_downloaded'
DOWNLOAD_CHOICE_LIST_DOWNLOADING = 'downloading'
DOWNLOAD_CHOICE_LIST_DOWNLOADED = 'downloaded'
DOWNLOAD_CHOICE_LIST_DICT = {
    None: DOWNLOAD_CHOICE_LIST_ALL,
    0: DOWNLOAD_CHOICE_LIST_NOT_DOWNLOAD,
    1: DOWNLOAD_CHOICE_LIST_DOWNLOADING,
    2: DOWNLOAD_CHOICE_LIST_DOWNLOADED,
}

DOWNLOAD_CHOICE = (DOWNLOAD_CHOICE_LIST_ALL, DOWNLOAD_CHOICE_LIST_DOWNLOADED,
                   DOWNLOAD_CHOICE_LIST_DOWNLOADING, DOWNLOAD_CHOICE_LIST_NOT_DOWNLOAD)

FOLLOWED_ACTION_LIST = 'list'
FOLLOWED_ACTION_MARK = 'mark'
FOLLOWED_CHOICE = (FOLLOWED_ACTION_LIST, FOLLOWED_ACTION_MARK)


# global Ctrl-C signal handler
def signal_handler(signal, frame):
    print_error('User aborted, quit')
signal.signal(signal.SIGINT, signal_handler)


# main function
def main():
    c = CommandParser()
    action = c.add_arg_group('action')

    sub_parser_add = action.add_sub_parser(ACTION_ADD, help='Subscribe bangumi.')
    sub_parser_add.add_argument('name', arg_type='+', required=True, help='Bangumi name to subscribe.')

    sub_parser_filter = action.add_sub_parser(ACTION_FILTER, help='Set bangumi fetch filter.')
    sub_parser_filter.add_argument('name', required=True, help='Bangumi name to set the filter.')
    sub_parser_filter.add_argument('--subtitle', arg_type='1', help='Subtitle group name, split by ",".')
    sub_parser_filter.add_argument('--include', arg_type='1',
                                   help='Filter by keywords which in the title, split by ",".')
    sub_parser_filter.add_argument('--exclude', arg_type='1',
                                   help='Filter by keywords which not int the title, split by ",".')
    # sub_parser_filter.add_argument('--remove', help='Remove subtitle group filter.')
    # sub_parser_filter.add_argument('--remove-all', help='Remove all the subtitle group filter.', mutex='--remove')

    sub_parser_del = action.add_sub_parser(ACTION_DELETE, help='Unsubscribe bangumi.')
    sub_parser_del.add_argument('--name', arg_type='+', mutex='--clear-all', help='Bangumi name to unsubscribe.')
    sub_parser_del.add_argument('--clear-all', help='Clear all the subscriptions.')
    sub_parser_del.add_argument('--batch', help='No confirmation.')

    sub_parser_fetch = action.add_sub_parser(ACTION_FETCH, help='Fetch a specific bangumi.')
    sub_parser_fetch.add_argument('name', help='Bangumi name to fetch.', required=True)
    sub_parser_fetch.add_argument('--not-ignore', help='Do not ignore the old bangumi detail rows (3 month ago).')

    sub_parser_update = action.add_sub_parser(ACTION_UPDATE, help='Update bangumi calendar and '
                                                                  'subscribed bangumi episode.')
    sub_parser_update.add_argument('--name', arg_type='+', help='Update specified bangumi.')
    sub_parser_update.add_argument('--download', help='Download the bangumi when updated.')
    sub_parser_update.add_argument('--not-ignore', help='Do not ignore the old bangumi detail rows (3 month ago).')

    sub_parser_cal = action.add_sub_parser(ACTION_CAL, help='Print bangumi calendar.')
    sub_parser_cal.add_argument('filter', default='today', choice=FILTER_CHOICES,
                                help='Calendar form filter %s.' % ', '.join(FILTER_CHOICES))
    sub_parser_cal.add_argument('--today', help='Show bangumi calendar for today.')
    sub_parser_cal.add_argument('--force-update', help='Get the newest bangumi calendar from dmhy.')
    sub_parser_cal.add_argument('--no-save', help='Do not save the bangumi data when force update.')

    sub_parser_config = action.add_sub_parser(ACTION_CONFIG, help='Config BGmi.')
    sub_parser_config.add_argument('name', help='Config name')
    sub_parser_config.add_argument('value', help='Config value')

    sub_parser_followed = action.add_sub_parser(ACTION_FOLLOWED, help='Subscribed bangumi manager.')
    sub_parser_followed.add_sub_parser('list', help='List subscribed bangumi.')
    followed_mark = sub_parser_followed.add_sub_parser('mark', help='Mark specific bangumi\'s episode.')
    followed_mark.add_argument('name', help='Bangumi name.', required=True)
    followed_mark.add_argument('episode', help='Bangumi episode.')

    sub_parser_mark = action.add_sub_parser(ACTION_MARK, help='Mark subscribed bangumi status.')
    sub_parser_mark.add_argument('name', help='Bangumi name.', required=True)
    sub_parser_mark.add_argument('episode', help='Bangumi episode.')

    sub_parser_download = action.add_sub_parser(ACTION_DOWNLOAD, help='Download manager.')
    download_list = sub_parser_download.add_sub_parser('list', help='List download queue.')
    download_list.add_argument('status', help='Download status: 0, 1, 2', choice=(0, 1, 2, None))

    download_mark = sub_parser_download.add_sub_parser('mark', help='Mark download status with a specific id.')
    download_mark.add_argument('id', help='Download id')
    download_mark.add_argument('status', help='Status will be marked', choice=(0, 1, 2))

    positional = c.add_arg_group('positional')
    positional.add_argument('install', help='Install xunlei-lixian for BGmi.')

    c.add_argument('-h / --help', help='Print help text.')
    c.add_argument('--version', help='Show the version of BGmi.')

    ret = c.parse_command()

    if ret.version:
        print_version()
        raise SystemExit

    if ret.positional.install == 'install':
        import bgmi.setup
        bgmi.setup.install()
        raise SystemExit

    elif ret.action == ACTION_ADD:
        add(ret)

    elif ret.action == ACTION_FILTER:
        filter_(ret)

    elif ret.action == ACTION_FETCH:
        bangumi_obj = Bangumi(name=ret.action.fetch.name)
        bangumi_obj.select_obj()

        followed_obj = Followed(bangumi_name=bangumi_obj.name)
        followed_obj.select_obj()

        if bangumi_obj:
            print_info('Fetch bangumi {0} ...'.format(bangumi_obj.name))
            _, data = get_maximum_episode(bangumi_obj,
                                          ignore_old_row=False if ret.action.fetch.not_ignore else True)
            if not data:
                print_warning('Nothing.')
            for i in data:
                print_success(i['title'])

        else:
            print_error('Bangumi {0} not exist'.format(ret.action.fetch.name))

    elif ret.action == ACTION_DELETE:
        delete(ret)

    elif ret.action == ACTION_UPDATE:
        update(ret)

    elif ret.action == ACTION_CAL:
        cal(ret)

    elif ret.action == ACTION_CONFIG:
        write_config(ret.action.config.name, ret.action.config.value)

    elif ret.action == ACTION_FOLLOWED:
        if not ret.action.followed == 'mark' and not ret.action.followed.list:
            c.print_help()
        else:
            followed(ret)

    elif ret.action == ACTION_MARK:
        mark(ret)

    elif ret.action == ACTION_DOWNLOAD:
        if ret.action.download in DOWNLOAD_ACTION:
            download_manager(ret)
        else:
            c.print_help()
    else:
        c.print_help()


def add(ret):
    # action add
    # add bangumi by a list of bangumi name
    if not Bangumi.get_all_bangumi():
        print_warning('No bangumi data in database, fetching...')
        update(ret)

    for bangumi in ret.action.add.name:
        bangumi_obj = Bangumi(name=bangumi)
        data = bangumi_obj.select(one=True, fields=['id', 'name', 'keyword'])
        if data:
            followed_obj = Followed(bangumi_name=data['name'], status=STATUS_FOLLOWED)
            followed_obj.select_obj()
            if not followed_obj or followed_obj.status == STATUS_NORMAL:
                if not followed_obj:
                    ret, _ = get_maximum_episode(bangumi_obj, subtitle=False)
                    followed_obj.episode = ret['episode']
                    followed_obj.save()
                else:
                    followed_obj.status = STATUS_FOLLOWED
                    followed_obj.save()
                print_success('{0} has been followed'.format(bangumi_obj))
            else:
                print_warning('{0} already followed'.format(bangumi_obj))

        else:
            print_error('{0} not found, please check the name'.format(bangumi))


def filter_(ret):
    bangumi_obj = Bangumi(name=ret.action.filter.name)
    bangumi_obj.select_obj()
    if not bangumi_obj:
        print_error('Bangumi {0} does not exist.'.format(bangumi_obj.name))

    followed_obj = Followed(bangumi_name=bangumi_obj.name)
    followed_obj.select_obj()

    if not followed_obj:
        print_error('Bangumi {0} has not subscribed, try \'bgmi add "{1}"\'.'.format(bangumi_obj.name,
                                                                                     bangumi_obj.name))

    subtitle = ret.action.filter.subtitle
    include = ret.action.filter.include
    exclude = ret.action.filter.exclude

    followed_filter_obj = Filter(bangumi_name=bangumi_obj.name)
    followed_filter_obj.select_obj()

    if not followed_filter_obj:
        followed_filter_obj.save()

    if subtitle is not None:
        subtitle = map(lambda s: s.strip(), subtitle.split(','))
        subtitle = filter(lambda s: True if s in bangumi_obj.subtitle_group.split(', ') else False, subtitle)
        subtitle = ', '.join(subtitle)
        followed_filter_obj.subtitle = subtitle

    if include is not None:
        followed_filter_obj.include = include

    if exclude is not None:
        followed_filter_obj.exclude = exclude

    followed_filter_obj.save()

    print_info('Usable subtitle group: {0}'.format(bangumi_obj.subtitle_group))
    print_success('Added subtitle group: {0}'.format(followed_filter_obj.subtitle))
    print_success('Include keywords: {0}'.format(followed_filter_obj.include))
    print_success('Exclude keywords: {0}'.format(followed_filter_obj.exclude))


def delete(ret):
    # action delete
    # just delete subscribed bangumi or clear all the subscribed bangumi
    if ret.action.delete.clear_all:
        if Followed.delete_followed(batch=ret.action.delete.batch):
            print_success('all subscriptions have been deleted')
        else:
            print_error('user canceled')
    elif ret.action.delete.name:
        for name in ret.action.delete.name:
            followed = Followed(bangumi_name=name)
            if followed.select():
                followed.delete()
                print_warning('Bangumi %s has been deleted' % name)
            else:
                print_error('Bangumi %s does not exist' % name, exit_=False)
    else:
        print_warning('Nothing has been done.')


def update(ret):
    ignore = False if ret.action.update.not_ignore else True
    print_info('marking bangumi status ...')
    now = int(time.time())
    for i in Followed.get_all_followed():
        if i['updated_time'] and int(i['updated_time'] + 86400) < now:
            followed_obj = Followed(bangumi_name=i['bangumi_name'])
            followed_obj.status = STATUS_FOLLOWED
            followed_obj.save()

    print_info('updating bangumi data ...')
    fetch(save=True, group_by_weekday=False)
    print_info('updating subscriptions ...')
    download_queue = []

    if ret.action.update.name is None:
        updated_bangumi_obj = Followed.get_all_followed()
    else:
        updated_bangumi_obj = []
        for i in ret.action.update.name:
            f = Followed(bangumi_name=i)
            f.select_obj()
            updated_bangumi_obj.append(f)

    for subscribe in updated_bangumi_obj:
        print_info('fetching %s ...' % subscribe['bangumi_name'])
        bangumi_obj = Bangumi(name=subscribe['bangumi_name'])
        bangumi_obj.select_obj()

        # filter by subtitle group
        if not bangumi_obj:
            print_error('The bangumi {0} you subscribed does not exists ..'.format(subscribe['bangumi_name']),
                        exit_=False)
            continue

        episode, all_episode_data = get_maximum_episode(bangumi=bangumi_obj, ignore_old_row=ignore)
        if episode.get('episode') > subscribe['episode']:
            episode_range = range(subscribe['episode'] + 1, episode.get('episode'))
            print_success('%s updated, episode: %d' % (subscribe['bangumi_name'], episode['episode']))
            _ = Followed(bangumi_name=subscribe['bangumi_name'])
            _.episode = episode['episode']
            _.status = STATUS_UPDATED
            _.updated_time = int(time.time())
            _.save()
            download_queue.append(episode)
            for i in episode_range:
                for epi in all_episode_data:
                    if epi['episode'] == i:
                        download_queue.append(epi)
                        break

    if ret.action.update and ret.action.update.download:
        download_prepare(download_queue)
        print_info('Re-downloading ...')
        download_prepare(Download.get_all_downloads(status=STATUS_NOT_DOWNLOAD))


def cal(ret):
    force = ret.action.cal.force_update
    save = not ret.action.cal.no_save
    today = ret.action.cal.today
    if ret.action.cal.filter == FILTER_CHOICE_TODAY:
        bangumi_calendar(force_update=force, today=True, save=save)
    elif ret.action.cal.filter == FILTER_CHOICE_FOLLOWED:
        bangumi_calendar(force_update=force, followed=True, today=today, save=save)
    else:
        # fallback
        bangumi_calendar(force_update=force, today=today, save=save)


def download_manager(ret):
    print_info('Download status value: Not Downloaded: 0 / Downloading: 1 / Downloaded: 2\n', indicator=False)

    if ret.action.download == DOWNLOAD_ACTION_LIST:
        status = ret.action.download.list.status
        status = int(status) if status is not None else None
        delegate = get_download_class(instance=False)
        delegate.download_status(status=status)

    elif ret.action.download == DOWNLOAD_ACTION_MARK:
        download_id = ret.action.download.mark.id
        status = ret.action.download.mark.status
        if not download_id or not status:
            print_error('No id or status specified.')
        download_obj = Download(_id=download_id)
        download_obj.select_obj()
        if not download_obj:
            print_error('Download object does not exist.')
        print_info('Download Object <{0} - {1}>, Status: {2}'.format(download_obj.name, download_obj.episode,
                                                                     download_obj.status))
        download_obj.status = status
        download_obj.save()
        print_success('Download status has been marked as {0}'.format(DOWNLOAD_CHOICE_LIST_DICT.get(int(status))))


def mark(ret):
    name = ret.action.mark.name
    episode = ret.action.mark.episode
    followed_obj = Followed(bangumi_name=name)
    followed_obj.select_obj()

    if not followed_obj:
        print_error('Subscribe <%s> does not exist.' % name)

    if episode:
        followed_obj.episode = episode
        followed_obj.save()
        print_success('%s has been mark as episode: %s' % (followed_obj, followed_obj.episode))
    else:
        print_info('%s, episode: %s' % (followed_obj, followed_obj.episode))


def followed(ret):
    if ret.action.followed == FOLLOWED_ACTION_MARK:
        print_warning('Warning: bgmi followed mark is deprecated, please use bgmi mark instead.')
        ret.action.mark = ret.action.followed.mark
        mark(ret)
    else:
        print_warning('Warning: bgmi followed list is deprecated, please use bgmi cal followed instead.')
        bangumi_calendar(force_update=False, followed=True, save=False)


def init_db(db_path):
    try:
        conn = sqlite3.connect(db_path)
        conn.execute(CREATE_TABLE_BANGUMI)
        conn.execute(CREATE_TABLE_FOLLOWED)
        conn.execute(CREATE_TABLE_DOWNLOAD)
        conn.execute(CREATE_TABLE_FOLLOWED_FILTER)
        conn.commit()
        conn.close()
    except sqlite3.OperationalError:
        print_error('Open database file failed, path %s is not writable.' % BGMI_PATH)


def setup():
    if not os.path.exists(BGMI_PATH):
        print_warning('BGMI_PATH %s does not exist, installing' % BGMI_PATH)
        from bgmi.setup import create_dir, install_crontab
        create_dir()
        if not platform.system() == 'Windows':
            # if not input('Do you want to install a crontab to auto-download bangumi?(Y/n): ') == 'n':
            install_crontab()

    # if not os.path.exists(DB_PATH):
    init_db(DB_PATH)
    main()


if __name__ == '__main__':
    setup()



# coding=utf-8
import sys
import unittest
from bgmi.command import CommandParser, NameSpace


class CommandTest(unittest.TestCase):

    def test_argument_name(self):
        c = CommandParser()
        group_1 = c.add_arg_group('test')
        group_1.add_argument('--id-aa--', arg_type='1')
        sys.argv = ['test.py', '--id-aa--', '123456']
        namespace_1 = c.parse_command()

        self.assertEqual(namespace_1.test.id_aa__, '123456')

    def test_parse_positional_argument(self):
        c = CommandParser()
        c.add_argument('method')
        c.add_argument('sub_method')
        sys.argv = ['test.py', 'method_test', 'sub_method_test']
        namespace_1 = c.parse_command()

        self.assertEqual(namespace_1.method, 'method_test')
        self.assertEqual(namespace_1.sub_method, 'sub_method_test')

    def test_sub_parser(self):
        c = CommandParser()
        group_1 = c.add_arg_group('test')
        sub_1 = group_1.add_sub_parser('sub_action')
        sub_2 = group_1.add_sub_parser('sub_action2')

        # conflict action name: --sub_action
        self.assertRaises(SystemExit, group_1.add_sub_parser, '--sub_action')

        sub_1.add_argument('--force')
        sub_1.add_argument('--id', arg_type='+')
        sub_2.add_argument('--verbose', arg_type='+')

        sys.argv = ['test.py', 'sub_action', '--force']
        namespace_1 = c.parse_command()
        self.assertEqual(namespace_1.test.sub_action.force, True)

        sys.argv = ['test.py', 'sub_action', '--force', '--sub_action2', '--verbose', '1']
        # unrecognized arguments: --sub_action2
        self.assertRaises(SystemExit, c.parse_command)

        sys.argv = ['test.py', 'sub_action2', '--force']
        # unrecognized arguments: --force
        self.assertRaises(SystemExit, c.parse_command)
        sys.argv = ['test.py', 'sub_action2', '--verbose', '666']
        namespace_1 = c.parse_command()
        self.assertEqual(namespace_1.test.sub_action2.verbose, ['666', ])

        d = CommandParser()
        group_2 = d.add_arg_group('action')
        s_1 = group_2.add_sub_parser('update')
        s_1.add_argument('subaction')
        s_2 = group_2.add_sub_parser('delete')
        s_2.add_argument('--subaction2')

        sys.argv = ['test.py', 'update', 'help']
        namespace_2 = d.parse_command()
        self.assertEqual(namespace_2.action.update.subaction, 'help')

        sys.argv = ['test.py', 'delete', '--subaction2']
        namespace_2 = d.parse_command()
        self.assertEqual(namespace_2.action.delete.subaction2, True)

    def test_parse_command(self):
        c = CommandParser()
        group_1 = c.add_arg_group('test')
        # positional argument
        group_1.add_argument('method')
        group_1.add_argument('-v', dest='verbose', default=False)
        group_1.add_argument('--id', dest='id', arg_type='1', default=False)
        group_1.add_argument('--id2', dest='id2', arg_type='+')
        group_1.add_argument('--id3', dest='id3', arg_type='+')

        sys.argv = ['test.py', 'method_test1']
        namespace_1 = c.parse_command()
        self.assertFalse(namespace_1.test.verbose)

        sys.argv = ['test.py', 'method_test2', '-v']
        namespace_2 = c.parse_command()
        self.assertTrue(namespace_2.test.verbose)

        sys.argv = ['test.py', 'method_test3', '-v', '--id', '1', '--id2', '1', '2']
        namespace_3 = c.parse_command()
        self.assertEqual(namespace_3.test.method, 'method_test3')
        self.assertEqual(namespace_3.test.id, '1')
        self.assertEqual(namespace_3.test.id2, ['1', '2'])

    def test_choice(self):
        c = CommandParser()
        g1 = c.add_arg_group('test')
        g1.add_argument('--method', arg_type='1', choice=('GET', 'POST', 'HEAD', ))
        sys.argv = ['test.py', '--method', 'PUT']
        self.assertRaises(SystemExit, c.parse_command)

        c = CommandParser()
        g2 = c.add_arg_group('test')
        sub_1 = g2.add_sub_parser('get_method')
        sub_1.add_argument('--get', arg_type='1', choice=('GET', 'HEAD', ), default='GET')

        sub_2 = g2.add_sub_parser('post_method')
        sub_2.add_argument('--post', arg_type='1', choice=('POST', 'PUT', ))

        sys.argv = ['test.py', 'get_method']
        namespace = c.parse_command()
        self.assertEqual(namespace.test.get_method.get, 'GET')

    def test_required(self):
        c = CommandParser()
        g1 = c.add_arg_group('test')
        g1.add_argument('method', required=True)
        g1.add_argument('--get', required=True)

        sys.argv = ['test.py']
        self.assertRaises(SystemExit, c.parse_command)

        sys.argv = ['test.py', 'GET']
        self.assertRaises(SystemExit, c.parse_command)

        sys.argv = ['test.py', 'GET', '--get']
        self.assertIsInstance(c.parse_command(), NameSpace)

    def test_position_arg_type(self):
        c = CommandParser()
        g1 = c.add_arg_group('test')
        g1.add_argument('position', arg_type='+')
        g1.add_argument('--verbose')
        sys.argv = ['test.py', 'AAA', 'BBB', 'CCC', '--verbose']
        namespace = c.parse_command()
        self.assertEqual(namespace.test.position, ['AAA', 'BBB', 'CCC'])

    def test_mutex_arg(self):
        # ordinary
        c = CommandParser()
        g1 = c.add_arg_group('test')
        g1.add_argument('--clear-all', mutex='--delete')
        g1.add_argument('--delete', arg_type='+')
        sys.argv = ['test.py', '--delete', 'AAA', '--clear-all']
        self.assertRaises(SystemExit, c.parse_command)

        # sub parser
        c = CommandParser()
        g2 = c.add_arg_group('test')
        sub = g2.add_sub_parser('sub')
        sub.add_argument('--clear-all', mutex='--delete')
        sub.add_argument('--delete', arg_type='+')
        sys.argv = ['test.py', 'sub', '--delete', 'AAA', '--clear-all']
        self.assertRaises(SystemExit, c.parse_command)

        c = CommandParser()
        g3 = c.add_arg_group('test')
        g3.add_argument('--clear-all-sub', mutex='--delete-sub')
        sub2 = g3.add_sub_parser('sub')
        sub2.add_argument('--delete-sub', arg_type='+')
        sys.argv = ['test.py', '--clear-all-sub', 'sub', '--delete-sub', 'AAA']
        self.assertIsInstance(c.parse_command(), NameSpace)


if __name__ == '__main__':
    unittest.main()
